@@ -1,32 +1,37 @@
-// Copyright (c) 2013, Kenton Varda <temporal@gmail.com>
-// All rights reserved.
+// Copyright (c) 2013-2014 Sandstorm Development Group, Inc. and contributors
+// Licensed under the MIT License:
 //
-// Redistribution and use in source and binary forms, with or without
-// modification, are permitted provided that the following conditions are met:
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to deal
+// in the Software without restriction, including without limitation the rights
+// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+// copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 //
-// 1. Redistributions of source code must retain the above copyright notice, this
-//    list of conditions and the following disclaimer.
-// 2. Redistributions in binary form must reproduce the above copyright notice,
-//    this list of conditions and the following disclaimer in the documentation
-//    and/or other materials provided with the distribution.
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 //
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
-// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
-// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
-// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
-// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
-// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
-// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+// THE SOFTWARE.
 
 #ifndef CAPNP_LIST_H_
 #define CAPNP_LIST_H_
 
+#if defined(__GNUC__) && !CAPNP_HEADER_WARNINGS
+#pragma GCC system_header
+#endif
+
 #include "layout.h"
 #include "orphan.h"
 #include <initializer_list>
+#ifdef KJ_STD_COMPAT
+#include <iterator>
+#endif  // KJ_STD_COMPAT
 
 namespace capnp {
 namespace _ {  // private
@@ -162,37 +167,25 @@ struct List<T, Kind::PRIMITIVE> {
 
   private:
     _::ListBuilder builder;
+    template <typename U, Kind K>
+    friend struct _::PointerHelpers;
     friend class Orphanage;
     template <typename U, Kind K>
     friend struct ToDynamic_;
   };
 
-private:
-  inline static _::ListBuilder initAsElementOf(
-      _::ListBuilder& builder, uint index, uint size) {
-    return builder.initListElement(
-        index * ELEMENTS, _::elementSizeForType<T>(), size * ELEMENTS);
-  }
-  inline static _::ListBuilder getAsElementOf(
-      _::ListBuilder& builder, uint index) {
-    return builder.getListElement(index * ELEMENTS, _::elementSizeForType<T>());
-  }
-  inline static _::ListReader getAsElementOf(
-      const _::ListReader& reader, uint index) {
-    return reader.getListElement(index * ELEMENTS, _::elementSizeForType<T>());
-  }
+  class Pipeline {};
 
-  inline static _::ListBuilder initAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, uint size) {
-    return builder.initListField(index, _::elementSizeForType<T>(), size * ELEMENTS);
+private:
+  inline static _::ListBuilder initPointer(_::PointerBuilder builder, uint size) {
+    return builder.initList(_::elementSizeForType<T>(), size * ELEMENTS);
   }
-  inline static _::ListBuilder getAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, const word* defaultValue) {
-    return builder.getListField(index, _::elementSizeForType<T>(), defaultValue);
+  inline static _::ListBuilder getFromPointer(_::PointerBuilder builder, const word* defaultValue) {
+    return builder.getList(_::elementSizeForType<T>(), defaultValue);
   }
-  inline static _::ListReader getAsFieldOf(
-      const _::StructReader& reader, WirePointerCount index, const word* defaultValue) {
-    return reader.getListField(index, _::elementSizeForType<T>(), defaultValue);
+  inline static _::ListReader getFromPointer(
+      const _::PointerReader& reader, const word* defaultValue) {
+    return reader.getList(_::elementSizeForType<T>(), defaultValue);
   }
 
   template <typename U, Kind k>
@@ -271,8 +264,7 @@ struct List<T, Kind::STRUCT> {
       // expanded under any circumstances.  We're just going to throw it away anyway, and
       // transferContentFrom() already carefully compares the struct sizes before transferring.
       builder.getStructElement(index * ELEMENTS).transferContentFrom(
-          orphan.builder.asStruct(_::StructSize(
-              0 * WORDS, 0 * POINTERS, _::FieldSize::VOID)));
+          orphan.builder.asStruct(_::StructSize(0 * WORDS, 0 * POINTERS)));
     }
     inline void setWithCaveats(uint index, const typename T::Reader& reader) {
       // Mostly behaves like you'd expect `set` to behave, but with a caveat originating from
@@ -296,37 +288,25 @@ struct List<T, Kind::STRUCT> {
 
   private:
     _::ListBuilder builder;
+    template <typename U, Kind K>
+    friend struct _::PointerHelpers;
     friend class Orphanage;
     template <typename U, Kind K>
     friend struct ToDynamic_;
   };
 
-private:
-  inline static _::ListBuilder initAsElementOf(
-      _::ListBuilder& builder, uint index, uint size) {
-    return builder.initStructListElement(
-        index * ELEMENTS, size * ELEMENTS, _::structSize<T>());
-  }
-  inline static _::ListBuilder getAsElementOf(
-      _::ListBuilder& builder, uint index) {
-    return builder.getStructListElement(index * ELEMENTS, _::structSize<T>());
-  }
-  inline static _::ListReader getAsElementOf(
-      const _::ListReader& reader, uint index) {
-    return reader.getListElement(index * ELEMENTS, _::FieldSize::INLINE_COMPOSITE);
-  }
+  class Pipeline {};
 
-  inline static _::ListBuilder initAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, uint size) {
-    return builder.initStructListField(index, size * ELEMENTS, _::structSize<T>());
+private:
+  inline static _::ListBuilder initPointer(_::PointerBuilder builder, uint size) {
+    return builder.initStructList(size * ELEMENTS, _::structSize<T>());
   }
-  inline static _::ListBuilder getAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, const word* defaultValue) {
-    return builder.getStructListField(index, _::structSize<T>(), defaultValue);
+  inline static _::ListBuilder getFromPointer(_::PointerBuilder builder, const word* defaultValue) {
+    return builder.getStructList(_::structSize<T>(), defaultValue);
   }
-  inline static _::ListReader getAsFieldOf(
-      const _::StructReader& reader, WirePointerCount index, const word* defaultValue) {
-    return reader.getListField(index, _::FieldSize::INLINE_COMPOSITE, defaultValue);
+  inline static _::ListReader getFromPointer(
+      const _::PointerReader& reader, const word* defaultValue) {
+    return reader.getList(ElementSize::INLINE_COMPOSITE, defaultValue);
   }
 
   template <typename U, Kind k>
@@ -351,7 +331,8 @@ struct List<List<T>, Kind::LIST> {
     inline uint size() const { return reader.size() / ELEMENTS; }
     inline typename List<T>::Reader operator[](uint index) const {
       KJ_IREQUIRE(index < size());
-      return typename List<T>::Reader(List<T>::getAsElementOf(reader, index));
+      return typename List<T>::Reader(
+          _::PointerHelpers<List<T>>::get(reader.getPointerElement(index * ELEMENTS)));
     }
 
     typedef _::IndexingIterator<const Reader, typename List<T>::Reader> Iterator;
@@ -383,15 +364,17 @@ struct List<List<T>, Kind::LIST> {
     inline uint size() const { return builder.size() / ELEMENTS; }
     inline typename List<T>::Builder operator[](uint index) {
       KJ_IREQUIRE(index < size());
-      return typename List<T>::Builder(List<T>::getAsElementOf(builder, index));
+      return typename List<T>::Builder(
+          _::PointerHelpers<List<T>>::get(builder.getPointerElement(index * ELEMENTS)));
     }
     inline typename List<T>::Builder init(uint index, uint size) {
       KJ_IREQUIRE(index < this->size());
-      return typename List<T>::Builder(List<T>::initAsElementOf(builder, index, size));
+      return typename List<T>::Builder(
+          _::PointerHelpers<List<T>>::init(builder.getPointerElement(index * ELEMENTS), size));
     }
     inline void set(uint index, typename List<T>::Reader value) {
       KJ_IREQUIRE(index < size());
-      builder.setListElement(index * ELEMENTS, value.reader);
+      builder.getPointerElement(index * ELEMENTS).setList(value.reader);
     }
     void set(uint index, std::initializer_list<ReaderFor<T>> value) {
       KJ_IREQUIRE(index < size());
@@ -403,11 +386,11 @@ struct List<List<T>, Kind::LIST> {
     }
     inline void adopt(uint index, Orphan<T>&& value) {
       KJ_IREQUIRE(index < size());
-      builder.adopt(index * ELEMENTS, kj::mv(value));
+      builder.getPointerElement(index * ELEMENTS).adopt(kj::mv(value));
     }
     inline Orphan<T> disown(uint index) {
       KJ_IREQUIRE(index < size());
-      return Orphan<T>(builder.disown(index * ELEMENTS));
+      return Orphan<T>(builder.getPointerElement(index * ELEMENTS).disown());
     }
 
     typedef _::IndexingIterator<Builder, typename List<T>::Builder> Iterator;
@@ -416,37 +399,25 @@ struct List<List<T>, Kind::LIST> {
 
   private:
     _::ListBuilder builder;
+    template <typename U, Kind K>
+    friend struct _::PointerHelpers;
     friend class Orphanage;
     template <typename U, Kind K>
     friend struct ToDynamic_;
   };
 
-private:
-  inline static _::ListBuilder initAsElementOf(
-      _::ListBuilder& builder, uint index, uint size) {
-    return builder.initListElement(
-        index * ELEMENTS, _::FieldSize::POINTER, size * ELEMENTS);
-  }
-  inline static _::ListBuilder getAsElementOf(
-      _::ListBuilder& builder, uint index) {
-    return builder.getListElement(index * ELEMENTS, _::FieldSize::POINTER);
-  }
-  inline static _::ListReader getAsElementOf(
-      const _::ListReader& reader, uint index) {
-    return reader.getListElement(index * ELEMENTS, _::FieldSize::POINTER);
-  }
+  class Pipeline {};
 
-  inline static _::ListBuilder initAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, uint size) {
-    return builder.initListField(index, _::FieldSize::POINTER, size * ELEMENTS);
+private:
+  inline static _::ListBuilder initPointer(_::PointerBuilder builder, uint size) {
+    return builder.initList(ElementSize::POINTER, size * ELEMENTS);
   }
-  inline static _::ListBuilder getAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, const word* defaultValue) {
-    return builder.getListField(index, _::FieldSize::POINTER, defaultValue);
+  inline static _::ListBuilder getFromPointer(_::PointerBuilder builder, const word* defaultValue) {
+    return builder.getList(ElementSize::POINTER, defaultValue);
   }
-  inline static _::ListReader getAsFieldOf(
-      const _::StructReader& reader, WirePointerCount index, const word* defaultValue) {
-    return reader.getListField(index, _::FieldSize::POINTER, defaultValue);
+  inline static _::ListReader getFromPointer(
+      const _::PointerReader& reader, const word* defaultValue) {
+    return reader.getList(ElementSize::POINTER, defaultValue);
   }
 
   template <typename U, Kind k>
@@ -469,7 +440,7 @@ struct List<T, Kind::BLOB> {
     inline uint size() const { return reader.size() / ELEMENTS; }
     inline typename T::Reader operator[](uint index) const {
       KJ_IREQUIRE(index < size());
-      return reader.getBlobElement<T>(index * ELEMENTS);
+      return reader.getPointerElement(index * ELEMENTS).template getBlob<T>(nullptr, 0 * BYTES);
     }
 
     typedef _::IndexingIterator<const Reader, typename T::Reader> Iterator;
@@ -501,23 +472,23 @@ struct List<T, Kind::BLOB> {
     inline uint size() const { return builder.size() / ELEMENTS; }
     inline typename T::Builder operator[](uint index) {
       KJ_IREQUIRE(index < size());
-      return builder.getBlobElement<T>(index * ELEMENTS);
+      return builder.getPointerElement(index * ELEMENTS).template getBlob<T>(nullptr, 0 * BYTES);
     }
     inline void set(uint index, typename T::Reader value) {
       KJ_IREQUIRE(index < size());
-      builder.setBlobElement<T>(index * ELEMENTS, value);
+      builder.getPointerElement(index * ELEMENTS).template setBlob<T>(value);
     }
     inline typename T::Builder init(uint index, uint size) {
       KJ_IREQUIRE(index < this->size());
-      return builder.initBlobElement<T>(index * ELEMENTS, size * BYTES);
+      return builder.getPointerElement(index * ELEMENTS).template initBlob<T>(size * BYTES);
     }
     inline void adopt(uint index, Orphan<T>&& value) {
       KJ_IREQUIRE(index < size());
-      builder.adopt(index * ELEMENTS, kj::mv(value));
+      builder.getPointerElement(index * ELEMENTS).adopt(kj::mv(value));
     }
     inline Orphan<T> disown(uint index) {
       KJ_IREQUIRE(index < size());
-      return Orphan<T>(builder.disown(index * ELEMENTS));
+      return Orphan<T>(builder.getPointerElement(index * ELEMENTS).disown());
     }
 
     typedef _::IndexingIterator<Builder, typename T::Builder> Iterator;
@@ -526,37 +497,25 @@ struct List<T, Kind::BLOB> {
 
   private:
     _::ListBuilder builder;
+    template <typename U, Kind K>
+    friend struct _::PointerHelpers;
     friend class Orphanage;
     template <typename U, Kind K>
     friend struct ToDynamic_;
   };
 
-private:
-  inline static _::ListBuilder initAsElementOf(
-      _::ListBuilder& builder, uint index, uint size) {
-    return builder.initListElement(
-        index * ELEMENTS, _::FieldSize::POINTER, size * ELEMENTS);
-  }
-  inline static _::ListBuilder getAsElementOf(
-      _::ListBuilder& builder, uint index) {
-    return builder.getListElement(index * ELEMENTS, _::FieldSize::POINTER);
-  }
-  inline static _::ListReader getAsElementOf(
-      const _::ListReader& reader, uint index) {
-    return reader.getListElement(index * ELEMENTS, _::FieldSize::POINTER);
-  }
+  class Pipeline {};
 
-  inline static _::ListBuilder initAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, uint size) {
-    return builder.initListField(index, _::FieldSize::POINTER, size * ELEMENTS);
+private:
+  inline static _::ListBuilder initPointer(_::PointerBuilder builder, uint size) {
+    return builder.initList(ElementSize::POINTER, size * ELEMENTS);
   }
-  inline static _::ListBuilder getAsFieldOf(
-      _::StructBuilder& builder, WirePointerCount index, const word* defaultValue) {
-    return builder.getListField(index, _::FieldSize::POINTER, defaultValue);
+  inline static _::ListBuilder getFromPointer(_::PointerBuilder builder, const word* defaultValue) {
+    return builder.getList(ElementSize::POINTER, defaultValue);
   }
-  inline static _::ListReader getAsFieldOf(
-      const _::StructReader& reader, WirePointerCount index, const word* defaultValue) {
-    return reader.getListField(index, _::FieldSize::POINTER, defaultValue);
+  inline static _::ListReader getFromPointer(
+      const _::PointerReader& reader, const word* defaultValue) {
+    return reader.getList(ElementSize::POINTER, defaultValue);
   }
 
   template <typename U, Kind k>
@@ -567,4 +526,14 @@ private:
 
 }  // namespace capnp
 
+#ifdef KJ_STD_COMPAT
+namespace std {
+
+template <typename Container, typename Element>
+struct iterator_traits<capnp::_::IndexingIterator<Container, Element>>
+      : public std::iterator<std::random_access_iterator_tag, Element, int> {};
+
+}  // namespace std
+#endif  // KJ_STD_COMPAT
+
 #endif  // CAPNP_LIST_H_
